#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# NuVo Grand Concerto Plugin for Indigo
# Developed by Jeremy Swancoat

import os
import sys
import time
import serial

# Note the "indigo" module is automatically imported and made available inside
# our global name space by the host process.

##################################################################################################
class Plugin(indigo.PluginBase):

	##############################################################################################
	# Required Methods
	##############################################################################################
	##############################################################################################
	def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs):
		# Call the base class's init method
		indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
		# Setting debug to True will show you verbose debugging information in the Indigo Event Log
		# LOTS of debugLog commands are placed throughout this routine.
		self.debug = None
		self.conn = None
		self.portEnabled = False
		self.commQueue = []
		self.zonedev = {}
			
	########################################
	def __del__(self):
		indigo.PluginBase.__del__(self)
		
	##############################################################################################
	# Non-Required Methods (but still defined by Indigo)
	##############################################################################################

	########################################
	def startup(self):
		# startup process for the serial device. Other than open the serial port, there should be none really. It just runs.
		self.debugLog(u"Startup Called")
		
		if self.conn is None:
			pass
		else:
			self.conn.close()		
			
		self.portEnabled = False
		serialUrl = self.getSerialPortUrl(self.pluginPrefs, u"serialport")
		self.debugLog("Serial Port URL is " + serialUrl)
		self.conn = self.openSerial(u"NuVo Grand Concerto", serialUrl, 57600, stopbits=1, timeout=0.1, writeTimeout=1)
		if self.conn is not None:
			self.portEnabled = True
			indigo.server.log("Serial Port Open at " + serialUrl)
	
	def shutdown(self):
		# close serial port here
		self.debugLog(u"Shutdown Called")
		self.conn.close()
		indigo.server.log("Serial Port Closed")


	
	def validateDeviceConfigUi(self, valuesDict, typeId, devId):
		# Validate that appropriate selections were made in device creation / modification dialogs.
		# Ensure that Zone is not already used
		self.debugLog("Validating Device")
		self.debugLog("Zone Selected: " + valuesDict["zoneselect"])
		###Logic to see if device is in use
		if valuesDict["zoneselect"] in self.zonedev:
			self.debugLog("Zone Already Assigned to Device")
			return(False, valuesDict)
		#Need to add logic to test if the device is already used.
		self.debugLog("Zone Available")	
		return (True, valuesDict)
		
	# This section creates entries and deletes entries in a dictionary of
	# Zx , dev in an effort help find devices simply by what zone they control.		
	def deviceStartComm(self, dev):
		self.debugLog("deviceStartComm called")
		self.zonedev[dev.pluginProps["zoneselect"]] = dev
		self.debugLog("Just added: " + dev.pluginProps["zoneselect"] + " to zonedev")
		self.debugLog("Calling Status: " + dev.pluginProps["zoneselect"] + "STATUS?")
		self.commQueue.append(dev.pluginProps["zoneselect"] + "STATUS?")
		
	def deviceStopComm(self, dev):
		del self.zonedev[dev.pluginProps["zoneselect"]]
		self.debugLog("Just deleted: " + dev.pluginProps["zoneselect"] + " from zonedev")
		
	def validatePrefsConfigUi(self, valuesDict):
		
		errorsDict = indigo.Dict()		
		self.debugLog("Validating Serial Port")
		self.validateSerialPortUi(valuesDict, errorsDict, u"serialport")
		
		if valuesDict[u'showDebugInfo'] == True:
			self.debug = True
		else:
			self.debug = False
				
		
		if len(errorsDict) > 0:
			return (False, valuesDict, errorsDict)
		return (True, valuesDict)

		
	def closedPrefsConfigUi(self, valuesDict, userCancelled):
		# Basically, when the Config box closes, we want another chance to open the serial port.
		# Since that's all the startup method does, just call it from here.
		self.debugLog("closedPrefsConfigUI() called")
		
		#Obviously, if the box closes by cancelling, then do nothing.
		if userCancelled:
			pass
		else:
			self.startup()
		
	def runConcurrentThread(self):
		## This is what runs as the plugin is simply 'running'. 
		## Basically, if a command comes in from the serial device, you read it, parse it and hand it off to
		## whatever device/or variable needs to have its state updated.
		## In the meantime, you have a queue set up which gets loaded with commands from devices which must be fed
		## to the serial interface to be enacted.
		try:
			while True:
				if self.portEnabled:
					#self.debugLog("Loop Running")
					sleeptime = 0.1
					fromgrandconcerto = 0
					fromgrandconcerto = self.conn.readline()
					if len(fromgrandconcerto) > 3:
						sleeptime = 0.01
						self.debugLog("Nuvo: " + fromgrandconcerto[:-2])  #removes last two characters (CR and LF)
						self.parseToServer(fromgrandconcerto[:-2])
					if len(self.commQueue) > 0:
						sleeptime = 0.01
						# Sleep is set slow to keep processor use down, but putting 18 commands through with
						# 0.1 sleeptime takes almost 2 seconds, so if there's commands in the queue, drop the sleeptime, and
						# bang them through really fast. Then reset at beginning of loop.
						self.debugLog(self.commQueue[0])
						sendcount = self.conn.write("*" + self.commQueue.pop(0) + "\r")	
				self.sleep(sleeptime)
		except self.StopThread:
			pass	
		
	def stopConcurrentThread(self):
		self.stopThread = True
		
	def menuGetSources(self, filter, valuesDict, typeID, devID):
		sourcelist = []
		sourcelist.append(("SRC1",self.pluginPrefs[u"SRC1"]))
		sourcelist.append(("SRC2",self.pluginPrefs[u"SRC2"]))
		sourcelist.append(("SRC3",self.pluginPrefs[u"SRC3"]))
		sourcelist.append(("SRC4",self.pluginPrefs[u"SRC4"]))
		sourcelist.append(("SRC5",self.pluginPrefs[u"SRC5"]))
		sourcelist.append(("SRC6",self.pluginPrefs[u"SRC6"]))
		return sourcelist
		
	def menuGetZones(self, filter, valuesDict, typeID, devID):
		zonelist = []
		for zonecode, servdev in self.zonedev.iteritems():
			zonelist.append((zonecode,servdev.name))
		return zonelist
		
	def parseToServer(self, fromgc):
		#Looks Messy, but it's really just a bunch of tests to see what kind of statement it is
		#and then handling it - start with simple statements, so we can just move on from there.
		#All responses start with a '#' character, so start by stripping the first character.
		fromgc = fromgc[1:]
		if fromgc == "ALLOFF":
			#If ALLOFF is called, no zones are specified etc.. you just have to tell the server everything it has is off
			for zonecode, servdev in self.zonedev.iteritems():
				self.debugLog("All Off - " + zonecode + "called")
				servdev.updateStateOnServer("onOffState", value="off")
		elif fromgc[0] == "Z":
			#Statements starting with Z are Zone statements. No Z and you can just stop and discard
			if "," in fromgc:
				#Some Zone statements are messy and complicated around config etc. The ones we're interested
				#in are set up as a comma-delimted list
				resplist = fromgc.split(",")
				zonecode=resplist[0]
				if zonecode in self.zonedev:
					#Now that we know it's a Zone statement... make sure it's a zone we even have a device for
					#If not, we can move on.
					self.debugLog(zonecode + " found in devices list")
					servdev = self.zonedev[zonecode]
					if resplist[1] == "OFF":
						#Off statements are simple and only have two elements in the list. Zone and off.
						servdev.updateStateOnServer("onOffState", value="off")
						indigo.server.log(servdev.name + " is Off")
					elif resplist[1] == "ON":
						#Statements for active zones have a bunch of info. Pick it out and send to server. Easy!
						#But first, save the current zone parameters, so you can compare and only put changes into the Event log.
						#This prevents the log from filling full of on/off states and source changes every time you change the volume.
						savedstate = {}
						savedinput = {}
						savedvolume = {}
						dev = self.zonedev[zonecode]
						savedstate[zonecode] = dev.states["onOffState"]
						savedinput[zonecode] = dev.states["Input"]
						savedvolume[zonecode] = dev.states["Volume"]
						if savedstate[zonecode] == "OFF":
							#If the zone is turning ON, and WAS OFF, then report everything...
							#This is probably the only time you report the zone was on too
							servdev.updateStateOnServer("onOffState", value="on")
							indigo.server.log(servdev.name + " is ON")
							servdev.updateStateOnServer("Input", value=resplist[2])
							indigo.server.log(servdev.name + " source is " + resplist[2])
							if resplist[3] == "MUTE":
								indigo.server.log(servdev.name + " is muted")
							else:
								calcvol = 79 - int(resplist[3][-2:])
								servdev.updateStateOnServer("Volume", value=calcvol)
								self.debugLog("Reported Volume is " + resplist[3])
								indigo.server.log(servdev.name + " volume is " + str(calcvol))
						else:
							#Now, test the paramters. If the source is unchanged, no need to restate. Same with Volume
							servdev.updateStateOnServer("onOffState", value="on")
							servdev.updateStateOnServer("Input", value=resplist[2])
							if savedinput[zonecode] == resplist[2]:
								#Now, just make the rest of the logic on the input. If the input is the same, report only volume changes.
								#if the input is changed, report only the input changes, and skip the volume.
								servdev.updateStateOnServer("onOffState", value="on")
								servdev.updateStateOnServer("Input", value=resplist[2])
								if resplist[3] == "MUTE":
									indigo.server.log(servdev.name + " is muted")
								else:
									calcvol = 79 - int(resplist[3][-2:])
									servdev.updateStateOnServer("Volume", value=calcvol)
									self.debugLog("Reported Volume is " + resplist[3])
									indigo.server.log(servdev.name + " volume is " + str(calcvol))
							else:
								servdev.updateStateOnServer("onOffState", value="on")
								servdev.updateStateOnServer("Input", value=resplist[2])
								indigo.server.log(servdev.name + " source is " + resplist[2])
								if resplist[3] == "MUTE":
									indigo.server.log(servdev.name + " is muted")
								else:
									calcvol = 79 - int(resplist[3][-2:])
									servdev.updateStateOnServer("Volume", value=calcvol)
									self.debugLog("Reported Volume is " + resplist[3])
									
						
				else:
					self.debugLog(zonecode + " is not yet assigned to a device")
			
	def actionControlDimmerRelay(self, action, dev):
		if action.deviceAction == indigo.kDeviceAction.TurnOn:
			indigo.server.log("Turn On " + dev.name)
			self.debugLog("Turn On " + dev.pluginProps["zoneselect"])
			self.commQueue.append(dev.pluginProps["zoneselect"] + "ON")
		elif action.deviceAction == indigo.kDeviceAction.TurnOff:
			indigo.server.log("Turn Off " + dev.name)
			self.debugLog("Turn Off " + dev.pluginProps["zoneselect"])
			self.commQueue.append(dev.pluginProps["zoneselect"] + "OFF")
		elif action.deviceAction == indigo.kDeviceAction.Toggle:
			indigo.server.log("Toggle " + dev.name)
			self.debugLog("Toggle " + dev.pluginProps["zoneselect"])
			self.commQueue.append(dev.pluginProps["zoneselect"] + "POWER")
			
	
	def setZoneVolume(self, pluginAction, dev):
		indigo.server.log("Set volume of " + dev.name + " to " + str(pluginAction.props.get(u"reqvolume")))
		try:
			newVolume = 79 - int(pluginAction.props.get(u"reqvolume"))
		except ValueError:
			indigo.server.log(u"Invalid Volume Level: " + pluginAction.props.get(u"reqvolume"))
			return
		
		self.debugLog("Setting Volume of " + dev.name + " to " + str(pluginAction.props.get(u"reqvolume")))	
		self.commQueue.append(dev.pluginProps["zoneselect"] + "VOL" + str(newVolume))	
		
	def setZoneSource(self, pluginAction, dev):

		indigo.server.log("Setting Source of " + dev.name + " to " + pluginAction.props.get(u"reqzone"))	
		self.commQueue.append(dev.pluginProps["zoneselect"] + pluginAction.props.get(u"reqzone"))
		
	def setAllOff(self, pluginAction):
		self.commQueue.append("ALLOFF")

	def selectFavorite(self, pluginAction, dev):
		#This code is still active but the XML that calls it is commented out.
		#This code works just fine, but for some reason the GC does not respond to the command
		#as it should. It simply returns '#?'. If I can get it to work, just uncomment the XML and
		#the action becomes available again.
		indigo.server.log("Setting Favorite of " + dev.name + " to " + pluginAction.props.get(u"reqfav"))	
		self.commQueue.append(dev.pluginProps["zoneselect"] + "FAV" + pluginAction.props.get(u"reqfav"))

	def sendTextAlert(self, pluginAction):
		
		self.debugLog("sendTextAlert called")
		mtype=pluginAction.props.get(u"msgtype")
		mdwell=pluginAction.props.get(u"msgdwell")
		message=str(pluginAction.props.get(u"messagetext"))
		message=self.substituteVariable(message)
		self.debugLog(message)
		for x in pluginAction.props.get(u"selectedzones"):
			self.debugLog("Sending Message Command: " + x + "MSG\"" + message + "\"," + mtype + "," + mdwell)
			self.commQueue.append(x + "MSG\"" + message + "\"," + mtype + "," + mdwell)
			
	def speakAnnouncement(self, pluginAction):
		
		self.debugLog("speakAnnouncement called")
		#First, save current zone parameters
		self.debugLog("Saving Zone Parameters")
		savedstate = {}
		savedinput = {}
		savedvolume = {}
		for zonecode in pluginAction.props.get(u"selectedzones"):
			dev = self.zonedev[zonecode]
			savedstate[zonecode] = dev.states["onOffState"]
			self.debugLog("Saved State: "+ zonecode + ": " + savedstate[zonecode])
			savedinput[zonecode] = dev.states["Input"]
			self.debugLog("Saved Input: "+ zonecode + ": " + savedinput[zonecode])
			savedvolume[zonecode] = dev.states["Volume"]
			self.debugLog("Saved Volume: "+ zonecode + ": " + str(int(savedvolume[zonecode])))
				
		#Now, go through all of the zones and switch them to the right input.
		self.debugLog("Setting to correct input")
		for zonecode in pluginAction.props.get(u"selectedzones"):
			self.commQueue.append(zonecode + pluginAction.props.get(u"annzone"))
			try:
				newVolume = 79 - int(pluginAction.props.get(u"annvolume"))
			except ValueError:
				indigo.server.log(u"Invalid Volume Level: " + pluginAction.props.get(u"annvolume"))
				return
			self.commQueue.append(zonecode + "VOL" + str(newVolume))
			self.commQueue.append(zonecode + "ON")
			
		
		#Make Announcement
		self.sleep(2)
		announcetext=pluginAction.props.get("announcement")
		self.debugLog("Ann Text: " + announcetext)
		announcetext=self.substituteVariable(announcetext)
		self.debugLog("After Sub, Ann Text: " + announcetext)
		self.debugLog("Speaking Now...")
		wait = not pluginAction.props.get(u"manualWait")
		waittime = 0
		if pluginAction.props.get(u"manualWait") == True:
			waittime = int(pluginAction.props.get(u"mantime"))
		indigo.server.speak(announcetext, waitUntilDone=wait)
		self.sleep(waittime)
		if pluginAction.props.get(u"airplaydelay") == True:
			self.sleep(2)
		
		
		#And finally, reset original parameters
		self.debugLog("Resetting Orig Zone Parameters")
		for zonecode in pluginAction.props.get(u"selectedzones"):
			self.commQueue.append(zonecode + savedstate[zonecode])
			self.commQueue.append(zonecode + savedinput[zonecode])
			origVolume = 79 - int(savedvolume[zonecode])
			self.commQueue.append(zonecode + "VOL" + str(origVolume))
			